@@ -46,6 +46,7 @@
 #endif
 
 #include <math.h>
+#include <string.h>
 #include "gtk-clutter-embed.h"
 #include "gtk-clutter-offscreen.h"
 #include "gtk-clutter-actor.h"
@@ -66,6 +67,10 @@
 #include <clutter/win32/clutter-win32.h>
 #endif
 
+#if defined(CLUTTER_WINDOWING_WAYLAND)
+#include <clutter/wayland/clutter-wayland.h>
+#endif
+
 #if defined(GDK_WINDOWING_X11)
 #include <gdk/gdkx.h>
 #endif
@@ -74,6 +79,10 @@
 #include <gdk/gdkwin32.h>
 #endif
 
+#if defined(GDK_WINDOWING_WAYLAND)
+#include <gdk/gdkwayland.h>
+#endif
+
 G_DEFINE_TYPE (GtkClutterEmbed, gtk_clutter_embed, GTK_TYPE_CONTAINER);
 
 #define GTK_CLUTTER_EMBED_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_CLUTTER_TYPE_EMBED, GtkClutterEmbedPrivate))
@@ -92,6 +101,11 @@ struct _GtkClutterEmbedPrivate
 
   guint geometry_changed : 1;
   guint use_layout_size : 1;
+
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+  struct wl_subcompositor *subcompositor;
+  struct wl_subsurface *subsurface;
+#endif
 };
 
 enum
@@ -251,6 +265,19 @@ gtk_clutter_filter_func (GdkXEvent *native_event,
   return GDK_FILTER_CONTINUE;
 }
 
+static gboolean
+gtk_clutter_embed_draw (GtkWidget *widget, cairo_t *cr)
+{
+#if defined(CLUTTER_WINDOWING_GDK)
+  GtkClutterEmbedPrivate *priv = GTK_CLUTTER_EMBED (widget)->priv;
+
+  if (clutter_check_windowing_backend (CLUTTER_WINDOWING_GDK))
+    clutter_stage_ensure_redraw (CLUTTER_STAGE (priv->stage));
+#endif
+
+  return GTK_WIDGET_CLASS (gtk_clutter_embed_parent_class)->draw (widget, cr);
+}
+
 static void
 gtk_clutter_embed_realize (GtkWidget *widget)
 {
@@ -363,6 +390,28 @@ gtk_clutter_embed_realize (GtkWidget *widget)
     }
 #endif
 
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+  if (priv->subcompositor)
+    {
+      GdkDisplay *display;
+      struct wl_surface *clutter_surface, *gtk_surface;
+      struct wl_compositor *compositor;
+
+      display = gtk_widget_get_display (widget);
+      compositor = gdk_wayland_display_get_wl_compositor (display);
+      clutter_surface = wl_compositor_create_surface (compositor);
+      gtk_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel(window));
+      clutter_wayland_stage_set_wl_surface (CLUTTER_STAGE (priv->stage),
+                                            clutter_surface);
+      priv->subsurface =
+              wl_subcompositor_get_subsurface (priv->subcompositor,
+                                               clutter_surface,
+                                               gtk_surface);
+      wl_subsurface_set_position (priv->subsurface, allocation.x, allocation.y);
+      wl_subsurface_set_desync (priv->subsurface);
+    }
+#endif
+
   clutter_actor_realize (priv->stage);
 
   if (gtk_widget_get_visible (widget))
@@ -533,6 +582,10 @@ gtk_clutter_embed_size_allocate (GtkWidget     *widget,
 	     draw before we get the ConfigureNotify response. */
 	  clutter_x11_handle_event ((XEvent *)&xevent);
 	}
+#endif
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+      if (priv->subsurface)
+        wl_subsurface_set_position (priv->subsurface, allocation->x, allocation->y);
 #endif
     }
 }
@@ -900,6 +953,7 @@ gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
 
   widget_class->style_updated = gtk_clutter_embed_style_updated;
   widget_class->size_allocate = gtk_clutter_embed_size_allocate;
+  widget_class->draw = gtk_clutter_embed_draw;
   widget_class->realize = gtk_clutter_embed_realize;
   widget_class->unrealize = gtk_clutter_embed_unrealize;
   widget_class->show = gtk_clutter_embed_show;
@@ -939,6 +993,39 @@ gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
   g_object_class_install_property (gobject_class, PROP_USE_LAYOUT_SIZE, pspec);
 }
 
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+static void
+registry_handle_global (void *data,
+                        struct wl_registry *registry,
+                        uint32_t name,
+                        const char *interface,
+                        uint32_t version)
+{
+  GtkClutterEmbed *embed = data;
+  GtkClutterEmbedPrivate *priv = embed->priv;
+
+  if (strcmp (interface, "wl_subcompositor") == 0)
+    {
+      priv->subcompositor = wl_registry_bind (registry,
+                                              name,
+                                              &wl_subcompositor_interface,
+                                              1);
+    }
+}
+
+static void
+registry_handle_global_remove (void *data,
+                               struct wl_registry *registry,
+                               uint32_t name)
+{
+}
+
+static const struct wl_registry_listener registry_listener = {
+  registry_handle_global,
+  registry_handle_global_remove
+};
+#endif
+
 static void
 gtk_clutter_embed_init (GtkClutterEmbed *embed)
 {
@@ -954,8 +1041,10 @@ gtk_clutter_embed_init (GtkClutterEmbed *embed)
   /* we accept key focus */
   gtk_widget_set_can_focus (widget, TRUE);
 
+  G_GNUC_BEGIN_IGNORE_DEPRECATIONS
   /* disable double-buffering: it's automatically provided by OpenGL */
   gtk_widget_set_double_buffered (widget, FALSE);
+  G_GNUC_END_IGNORE_DEPRECATIONS
 
   /* we own the whole drawing of this widget, including the background */
   gtk_widget_set_app_paintable (widget, TRUE);
@@ -987,6 +1076,25 @@ gtk_clutter_embed_init (GtkClutterEmbed *embed)
     g_signal_connect (priv->stage,
                       "queue-relayout", G_CALLBACK (on_stage_queue_relayout),
                       embed);
+
+
+#if defined(GDK_WINDOWING_WAYLAND) && defined(CLUTTER_WINDOWING_WAYLAND)
+  {
+    GdkDisplay *gdk_display = gtk_widget_get_display (widget);
+    if (clutter_check_windowing_backend (CLUTTER_WINDOWING_WAYLAND) &&
+        GDK_IS_WAYLAND_DISPLAY (gdk_display))
+      {
+        struct wl_display *display;
+        struct wl_registry *registry;
+
+        display = gdk_wayland_display_get_wl_display (gdk_display);
+        registry = wl_display_get_registry (display);
+        wl_registry_add_listener (registry, &registry_listener, embed);
+
+        wl_display_roundtrip (display);
+      }
+  }
+#endif
 }
 
 /**
